//Fragment Shader
#version 440

// Uniforms:
layout(location = 0) uniform mat4 u_ModelMatrix;

// SceneUBO:
//THIS NEEDS TO MATCH SceneUBO.h!
const int LIGHT_COUNT = 50;
const int SHADOWMAP_COUNT = 3;
struct Light {
    vec4 position;
	vec4 direction;
    vec4 color;
	int type;

    float constant;
    float linear;
    float quadratic;
};
layout(std140, binding = 0) uniform Scene {
	mat4 u_ViewMatrix;
	mat4 u_ProjectMatrix;

	mat4 lightViewMatrices[4];
	mat4 lightOrthoProjMatrices[4];
	vec4 cascadeFarZ;				//we use a vec4 here casue glsl wants to pad each element of a float array up to a vec4.
	vec4 u_sunDirection;
	vec4 u_sunColor;

	vec4 u_Iambient;		//ambient illumination color
	int u_LightCount;

	Light u_Lights[LIGHT_COUNT];
};

//Varying:
layout(location=0) out vec4 fragColor;
layout(location=1) in vec2 v_UV;
layout(location=2) in vec3 v_Normal;
layout(location=3) in mat3 TBN;
layout(location=6) in vec4[SHADOWMAP_COUNT] v_modelLightViewVertexPos;
layout(location=9) in vec3 v_WorldPos;

//TEXTURES:
layout(binding=0) uniform sampler2D albedoMap;
layout(binding=1) uniform sampler2D normalMap;
layout(binding=2) uniform sampler2D metallicMap;
layout(binding=3) uniform sampler2D roughnessMap;
layout(binding=4) uniform sampler2D aoMap;

layout(binding=13) uniform sampler2DShadow shadowmap_0;
layout(binding=14) uniform sampler2DShadow shadowmap_1;
layout(binding=15) uniform sampler2DShadow shadowmap_2;


float sampleShadowPCFBlur(sampler2DShadow map, vec2 location, int radius, float bias, float depth){
	float shadow = 0.0f;
	vec2 texelSize = 1.0f / textureSize(map, 0);

	for(int x=-radius; x<=radius; x++){
		for(int y=-radius; y<=radius; y++){

			//When we sample a sampler2DShadow PCf should be done automatically. However, we need to make sure that we have the texture's compare func & mode configured properly
			float temp = texture(map, vec3(location + vec2(x,y)*texelSize, depth-bias)).r;

			//now compare the distance between the frag and sun (depth) with the value stored in the shadow map
			shadow += temp;//depth - bias > temp ? 0.1f : 1.0f;
		}
	}

	return shadow / ((2*radius+1)*(2*radius+1));
}




float getShadow(){
	//glFragCoord is only between 0 and 1. This should recover the real distance to the camera
	float originalZ = gl_FragCoord.z / gl_FragCoord.w;

	//check which texture we want to sample
	int index = 4;
	for (int i=0; i<SHADOWMAP_COUNT; i++) {
		if ( abs(originalZ) < cascadeFarZ[i] ) {
			index = i;
			break;
		}
	}

	//since multiple fragments mights need to share one pixel in the shadow map it can't be 100% perfect.
	//The result is a grainy pattern on surfaces
	//Especially on surfaces that are on a steep angle from the light source this is noticeable.
	//We use the dot product to create a bias based on the angle. We use this to be more leniant on steep angles.
	float surfaceLightDot = dot(v_Normal, u_sunDirection.xyz * (-1));
	float maxBias = 0.0055f;
	float minBias = 0.0012f;
	float bias = max(maxBias * (1.0f - abs(surfaceLightDot)), minBias);


	//screen coord (-1 to 1) to tex coords (0 to 1)
	vec3 texCoords = v_modelLightViewVertexPos[index].xyz * 0.5 + 0.5;

	float depth;
	if(index == 0){
		return sampleShadowPCFBlur(shadowmap_0, texCoords.xy, 1, bias, texCoords.z);
	} else if (index == 1) {
		return sampleShadowPCFBlur(shadowmap_1, texCoords.xy, 1, bias, texCoords.z);
	} else if (index == 2) {
		return sampleShadowPCFBlur(shadowmap_2, texCoords.xy, 1, bias, texCoords.z);
	}

	//This should never happen, but just in case let's just pretend that this surface is not in shadow.
	return 1.0f;
}

vec3 getShadowDEBUG(){
	vec3 result = abs(v_Normal);

	//glFragCoord is only between 0 and 1. This should recover the real distance to the camera
	float originalZ = gl_FragCoord.z / gl_FragCoord.w;
	int index = 4;
	for (int i=0; i<SHADOWMAP_COUNT; i++) {
		if ( abs(originalZ) < cascadeFarZ[i] ) {
			index = i;
			break;
		}
	}



	//screen coord (-1 to 1) to tex coords (0 to 1)
	vec3 texCoords = v_modelLightViewVertexPos[index].xyz * 0.5 + 0.5;

	if(texCoords.x <= 0.0){
		return vec3(1.0f, 0.0f, 0.0f);
	} else if(texCoords.y <= 0.0){
		return vec3(0.0f, 1.0f, 0.0f);
	}


	float depth;
	if(index == 0){
		//depth = texture(shadowmap_0, texCoords.xy).r;
		//result = vec3(1.0f,0.0f,0.0f);
	} else if (index == 1) {
		//depth = texture(shadowmap_1, texCoords.xy).r;
		//result = vec3(0.0f,1.0f,0.0f);
	} else if (index == 2) {
		//depth = texture(shadowmap_2, texCoords.xy).r;
		//result = vec3(0.0f,0.0f,1.0f);
	}

	//now compare the distance between the frag and the sun (texcoords.z) with the distance from the shadowmap
	if(texCoords.z-0.0005 < depth ) {
		//we are in sun
		return result;
	}

	return result * 0.3f;	//this should be 0
}

vec3 transformNormal(vec3 n){
	// right now we have two normals: the fragment normal and the texture normal. The problem is that the texture normal always points in
	// the z direction, regardless of where exactly the fragment is. We need to combine them both to get the normals in world space
	
	n = n * 2.0 - 1.0;   // 0 to 1  to  -1 to 1
	n = normalize(TBN * n); 
	return n;
}

// Converts a color from sRGB gamma to linear light gamma
vec3 toLinear(vec3 sRGB)
{
    bvec3 cutoff = lessThan(sRGB, vec3(0.04045));
    vec3 higher = pow((sRGB + vec3(0.055))/vec3(1.055), vec3(2.4));
    vec3 lower = sRGB/vec3(12.92);

    return mix(higher, lower, cutoff);
}
// Converts a color from linear light gamma to sRGB gamma
vec3 fromLinear(vec3 linearRGB)
{
    bvec3 cutoff = lessThan(linearRGB, vec3(0.0031308));
    vec3 higher = vec3(1.055)*pow(linearRGB, vec3(1.0/2.4)) - vec3(0.055);
    vec3 lower = linearRGB * vec3(12.92);

    return mix(higher, lower, cutoff);
}


//Fresnel Function F:  ratio between specular and diffuse reflection
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(max(1.0 - cosTheta, 0.0), 5.0);
} 


const float PI = 3.14159265359;
//Normal Distribution Function D: approximates the amount the surface's microfacets are aligned to the halfway vector
float DistributionGGX(vec3 N, vec3 H, float roughness)
{
    float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(N, H), 0.0);
    float NdotH2 = NdotH*NdotH;
	
    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;
	
    return num / denom;
}

//used by GeometrySmith
float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;
	
    return num / denom;
}

//Geometry Function G: describes the self-shadowing property of the microfacets.
//When a surface is relatively rough, the surface's microfacets can overshadow other microfacets reducing the light the surface reflects
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
    float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}



vec3 lightRadiance(vec3 L, vec3 N, vec3 V, float attenuation, vec3 F0, vec3 lightColor, vec3 albedo, float metallic, float roughness) {
	vec3 H = normalize(V+L);

	vec3 radiance = lightColor * attenuation;


	//Cook Torrence BRDF
	float D = 	DistributionGGX(N,H,roughness);
	float G = 	GeometrySmith(N,V,L, roughness);
	vec3 F =	fresnelSchlick(max(dot(H,V), 0.0), F0);

	vec3 kS = F;	//specular
	vec3 kD = vec3(1.0) - kS;	//diffuse
	kD *= 1.0-metallic;

	vec3 numerator = D*G*F;
	float denominator = 4.0 * max(dot(N, V), 0.0) * max(dot(N, L), 0.0);
	vec3 specular = numerator / max(denominator, 0.0001);	//don't divide by 0

	// outgoing radiance Lo
    float NdotL = max(dot(N, L), 0.0); 
	vec3 l = (kD * albedo / PI + specular) * radiance * NdotL;

	return l;
}

float calculateAttenuation(float distance, float constant, float linear, float quadratic){
	// Physically correct is the standard [0,0,1] but it can now be set again via the PointLight-Node
	return 1.0 / (constant + linear * distance + quadratic * distance * distance);
}

void main() {


    vec3 albedo     = toLinear(texture(albedoMap, v_UV).rgb);					//convert srgb albedo to linear color space.
    float metallic  = texture(metallicMap, v_UV).r;
    float roughness = texture(roughnessMap, v_UV).r;
	roughness = mix(0.1, 1.0, roughness);
    float ao        = texture(aoMap, v_UV).r;

	vec3 camPos = inverse(u_ViewMatrix)[3].xyz;

    vec3 N = transformNormal(texture(normalMap, v_UV).rgb);	//normal in world space
	vec3 V = normalize(camPos - v_WorldPos);


	//Surface reflection at 0 incidence (how much the surface reflects when directly looking at it)
	//needed for fresnel
    vec3 F0 = mix(vec3(0.04), albedo, metallic);



	vec3 light = vec3(0.0);


	//sun light
	light += lightRadiance(((-1.0)*u_sunDirection.xyz), N, V, 1.0, F0, u_sunColor.rgb, albedo, metallic, roughness) * getShadow();


	// + light of every element in the lights array
	for(int i=0; i<u_LightCount; i++){
		Light light_source = u_Lights[i];

		vec3 L=vec3(0.0f);
		float attenuation = 1.0f;

		switch(light_source.type) {
			case 0:
				//POINT_LIGHT
				L = light_source.position.xyz - v_WorldPos;
				attenuation = calculateAttenuation(length(L), light_source.constant, light_source.linear, light_source.quadratic);
				L = normalize(L);
				break;
			case 1:
				//DIRECTIONAL LIGHT
				L = normalize((-1.0)*light_source.direction.xyz);
				break;
		}

		light += lightRadiance(L, N, V, attenuation, F0, light_source.color.rgb, albedo, metallic, roughness);
	}


	//light end

	vec3 ambient = u_Iambient.rgb * albedo * ao;

	vec3 color = ambient + light;
	//color has now a HDR result


	//going back to low dynamic range for output:
    color = color / (color + vec3(1.0));
    color = pow(color, vec3(1.0/2.2));  


	fragColor = vec4(color, 1.0);
}