// Very very basic vertex shader
#version 440

// Attributes (per vertex information)
layout(location = 0) in vec3 a_Position;
layout(location = 1) in vec3 a_Normal;
layout(location = 2) in vec3 a_Tangent;
layout(location = 3) in vec3 a_Bitangent;
layout(location = 4) in vec2 a_UV;
layout(location = 5) in ivec4 a_BoneIds;
layout(location = 6) in vec4 a_BoneWeights;

//the following 2 values NEED TO MATCH Bone.h
const int MAX_BONES = 20;
// in order to change MAX_BONES_PER_VERTEX the type of a_BoneIds needs to be changed aswell
const int MAX_BONES_PER_VERTEX = 4;

// Uniforms:
layout(location = 0) uniform mat4 u_ModelMatrix;
// Location 1 is used for shadowmap
layout(location = 2) uniform bool u_isRigged;
layout(location = 3) uniform mat4 u_BoneTransforms[MAX_BONES];
layout(location = 100) uniform float u_opacity;
layout(location = 101) uniform vec4 u_modulateColor;


// SceneUBO:
//THIS NEEDS TO MATCH SceneUBO.h!
const int LIGHT_COUNT = 50;
const int SHADOWMAP_COUNT = 3;
struct Light {
    vec4 position;
	vec4 direction;
    vec4 color;
	int type;

    float constant;
    float linear;
    float quadratic;
};
layout(std140, binding = 0) uniform Scene {
	mat4 u_ViewMatrix;
	mat4 u_ProjectMatrix;

	mat4 lightViewMatrices[4];
	mat4 lightOrthoProjMatrices[4];
	vec4 cascadeFarZ;				//we use a vec4 here casue glsl wants to pad each element of a float array up to a vec4.
	vec4 u_sunDirection;
	vec4 u_sunColor;

	vec4 u_Iambient;		//ambient illumination color
	int u_LightCount;

	Light u_Lights[LIGHT_COUNT];
};

//Varying:
// 0 is color out
layout(location=1) out vec2 v_UV;
layout(location=2) out vec3 v_Normal;
layout(location=3) out mat3 v_TBN;
layout(location=6) out vec4[SHADOWMAP_COUNT] v_modelLightViewVertexPos;
layout(location=9) out vec3 v_WorldPos;
void main(){
	//mat4 mv_Matrix = u_ViewMatrix * u_ModelMatrix;
	//mat4 mvp_Matrix = u_ProjectMatrix * mv_Matrix;
    mat4 mvp_Matrix = u_ProjectMatrix * u_ModelMatrix;

	vec4 localPosition = vec4(a_Position, 1.0);
	vec4 localNormal = vec4(a_Normal, 0.0);
	vec4 localTangent = vec4(a_Tangent, 0.0);
	vec4 localBitangent = vec4(a_Bitangent, 0.0);
	if (u_isRigged) {
		localPosition = vec4(0.0);
		localNormal = vec4(0.0);
		localTangent = vec4(0.0);
		localBitangent = vec4(0.0);
		for (int i = 0; i < MAX_BONES_PER_VERTEX; i++) {
			if (a_BoneIds[i] < 0 || a_BoneIds[i] >= MAX_BONES) {
				break; 	// Invalid BoneId, that means the previous one was the last one
			}
			vec4 partPosition = u_BoneTransforms[a_BoneIds[i]] * vec4(a_Position, 1.0);
			localPosition += partPosition * a_BoneWeights[i];
			vec4 partNormal = u_BoneTransforms[a_BoneIds[i]] * vec4(a_Normal, 0.0);
			localNormal += partNormal * a_BoneWeights[i];
			vec4 partTangent = u_BoneTransforms[a_BoneIds[i]] * vec4(a_Tangent, 0.0);
			localTangent += partTangent * a_BoneWeights[i];
			vec4 partBitangent = u_BoneTransforms[a_BoneIds[i]] * vec4(a_Bitangent, 0.0);
			localBitangent += partBitangent * a_BoneWeights[i];
		}
	}

	gl_Position = mvp_Matrix * localPosition;

	// move stuff from attribute to varying to hand it to the fragment stage
	v_UV = a_UV;
	// The multiplication with the inverse ModelMatrix fixes the normal if non-uniform scaling is applied!
	// ToDo: The calculation should be done in C++ since the inverse is a complicated operation!
	mat4 normalModelMatrix = transpose(inverse(u_ModelMatrix));
	v_Normal = normalize(vec3(normalModelMatrix * vec4(localNormal.xyz, 0.0)));

	//TBN matrix used for normal mapping (left this here, if the other calculations don't work, maybe those will)
	//vec3 T = normalize(vec3(u_ModelMatrix * vec4(localTangent,   0.0)));	//TODO: this needs bone animations
	//vec3 B = normalize(vec3(u_ModelMatrix * vec4(a_Bitangent, 0.0)));	//TODO: this needs bone animations
	//vec3 N = normalize(vec3(u_ModelMatrix * vec4(a_Normal,    0.0)));	//TODO: this needs bone animations
	vec3 T = normalize(vec3(normalModelMatrix * vec4(localTangent.xyz, 0.0)));
	vec3 B = normalize(vec3(normalModelMatrix * vec4(localBitangent.xyz, 0.0)));
	vec3 N = v_Normal;
	v_TBN = mat3(T, B, N);

	v_WorldPos = (u_ModelMatrix * localPosition).xyz;

	//transfer vertex pos into light space so we can sample the shadow maps
	for (int i = 0 ; i < SHADOWMAP_COUNT ; i++) {
		v_modelLightViewVertexPos[i] = lightOrthoProjMatrices[i] * lightViewMatrices[i] * u_ModelMatrix * localPosition;
	}
}